from .cortexxdr import CortexXDR_pipeline
# TODO: add all pipelines that should be exposed to the user of your backend in the import statement above.

pipelines = {
    "CortexXDR_pipeline": CortexXDR_pipeline,   # TODO: adapt identifier to something approproiate
}